/*
 * Red-Unida IRC Daemon
 * Copyright (C) 2004-2005 Micael Fernandez
 *
 *
 * Basado sobre el codigo de Jarkko Oikarinen, UnderNet y IRC-Hispano.
 *
 * Este programa es gratis, puedes redistribuirlo y modificarlo bajo los
 * terminos de la GNU.
 *
 */

#include "sys.h"
#include "h.h"
#include "s_debug.h"
#include "numeric.h"
#include "errores.h"
#include "sprintf_irc.h"
#include "m_watch.h"

RCSTAG_CC("$Id: errores.c,v 2.0 2005/02/18 00:26:42 Oscuro Exp $");

typedef struct {
  int num_val;
  char *num_form;
} Numeric;

/* *INDENT-OFF* */

static Numeric local_replies[] = {
/* 000 */
    {0, (char *)NULL},
/* 001 */
    {RPL_WELCOME, ":Bienvenido a %s IRC Networks %s"},
/* 002 */
    {RPL_YOURHOST, ":Estas en el servidor %s, bajo version %s"},
/* 003 */
    {RPL_CREATED, ":Este servidor ha sido creado el %s"},

/* 004 */
#if defined(BDD_VIP)
    {RPL_MYINFO, "%s %s acdeghijkoprswxABRS bciklmnopstrvAKMR"},
#else
    {RPL_MYINFO, "%s %s acdeghijkoprswABRS bciklmnopstrvAKMR"}, 
#endif /* BDD_VIP */

/* 005 */
    {RPL_ISUPPORT, "%s :son soportados en este servidor"},
/* 006 */
    {0, (char *)NULL},
/* 007 */
    {0, (char *)NULL},
/* 008 */
    {RPL_SNOMASK, "%d :: Server notice mask (%#x)"},
/* 009 */
    {RPL_STATMEMTOT, "%u %u :Bytes Blocks"},
/* 010 */
#if defined(MEMSIZESTATS)
    {RPL_STATMEM, "%u %u %s %u"},
#else
    {RPL_STATMEM, "%u %u %s"},
#endif
/* 011 */
    {0, (char *)NULL},
/* 012 */
    {0, (char *)NULL},
/* 013 */
    {0, (char *)NULL},
/* 014 */
    {0, (char *)NULL},
/* 015 */
    {RPL_MAP, ":%s%s%s (%s) %s [%i clientes - %i.%i%%]"},
/* 016 */
    {RPL_MAPMORE, ":%s%s --> *mas*"},
/* 017 */
    {RPL_MAPEND, ":Fin del /MAP"},
    
    {0, (char *)NULL}
};

static Numeric numeric_errors[] = {
/* 401 */
    {ERR_NOSUCHNICK, "%s :No encuentro el nick"},
/* 402 */
    {ERR_NOSUCHSERVER, "%s :No encuentro el servidor"},
/* 403 */
    {ERR_NOSUCHCHANNEL, "%s :No encuentro el canal"},
/* 404 */
    {ERR_CANNOTSENDTOCHAN, "%s :No puedes enviar texto al canal"},
/* 405 */
    {ERR_TOOMANYCHANNELS, "%s :Has sobre pasado el limite de canales"},
/* 406 */
    {ERR_WASNOSUCHNICK, "%s :No encuentro nick"},
/* 407 */
    {ERR_TOOMANYTARGETS, "%s :Duplicate recipients. No message delivered"},
/* 408 */
    {0, (char *)NULL},
/* 409 */
    {ERR_NOORIGIN, ":No has especificado el origen"},
/* 410 */
    {0, (char *)NULL},
/* 411 */
    {ERR_NORECIPIENT, ":No has especificado el recibidor (%s)"},
/* 412 */
    {ERR_NOTEXTTOSEND, ":No hay texto para enviar"},
/* 413 */
    {ERR_NOTOPLEVEL, "%s :No toplevel domain specified"},
/* 414 */
    {ERR_WILDTOPLEVEL, "%s :Wildcard in toplevel Domain"},
/* 415 */
    {0, (char *)NULL},
/* 416 */
    {ERR_QUERYTOOLONG, "%s :Demasiadas lineas salientes, moderate en el query"},
/* 417 */
    {0, (char *)NULL},
/* 418 */
    {0, (char *)NULL},
/* 419 */
    {0, (char *)NULL},
/* 420 */
    {ERR_INVALIDBANMASK, "%s :Mascara de ban invalida: %s"},
/* 421 */
    {ERR_UNKNOWNCOMMAND, "%s :Comando desconocido"},
/* 422 */
    {ERR_NOMOTD, ":No hay mensaje del dia"},
/* 423 */
    {ERR_NOADMININFO, "%s :No hay informacion administrativa"},
/* 424 */
    {0, (char *)NULL},
/* 425 */
    {0, (char *)NULL},
/* 426 */
    {0, (char *)NULL},
/* 427 */
    {0, (char *)NULL},
/* 428 */
    {0, (char *)NULL},
/* 429 */
    {0, (char *)NULL},
/* 430 */
    {0, (char *)NULL},
/* 431 */
    {ERR_NONICKNAMEGIVEN, ":No hay nick"},
/* 432 */
    {ERR_INVALIDNICKNAME, "%s :Nick invalido"},
/* 433 */
    {ERR_NICKNAMEINUSE, "%s :Este nick esta siendo usado o esta registrado "
                        "(falta o no es correcto el password)"},
/* 434 */
    {0, (char *)NULL},
/* 435 */
    {0, (char *)NULL},
/* 436 */
    {ERR_NICKCOLLISION, "%s :Nick collision KILL"},
/* 437 */
    {ERR_BANNICKCHANGE, "%s :No te puedes cambiar de nick estando baneado en un canal"},
/* 438 */
    {ERR_NICKTOOFAST, "%s :Cambios de nick demasiado rapidos. Espera %d segundos"},
/* 439 */
    {ERR_TARGETTOOFAST, "%s :Objectivo demasiado rapido. Espera %d segundos"},
/* 440 */
    {0, (char *)NULL},
/* 441 */
    {ERR_USERNOTINCHANNEL, "%s %s :No esta en el canal"},
/* 442 */
    {ERR_NOTONCHANNEL, "%s :No estas en el canal"},
/* 443 */
    {ERR_USERONCHANNEL, "%s %s :Ya esta en el canal"},
/* 444 */
    {0, (char *)NULL},
/* 445 */
    {0, (char *)NULL},
/* 446 */
    {0, (char *)NULL},
/* 447 */
    {0, (char *)NULL},
/* 448 */
    {0, (char *)NULL},
/* 449 */
    {0, (char *)NULL},
/* 450 */
    {0, (char *)NULL},
/* 451 */
    {ERR_NOTREGISTERED, ":No estas registrado"},
/* 452 */
    {0, (char *)NULL},
/* 453 */
    {0, (char *)NULL},
/* 454 */
    {0, (char *)NULL},
/* 455 */
    {0, (char *)NULL},
/* 456 */
    {0, (char *)NULL},
/* 457 */
    {0, (char *)NULL},
/* 458 */
    {0, (char *)NULL},
/* 459 */
    {0, (char *)NULL},
/* 460 */
    {0, (char *)NULL},
/* 461 */
    {ERR_NEEDMOREPARAMS, "%s :Falta algun parametro"},
/* 462 */
    {ERR_ALREADYREGISTRED, ":Ya estas registrado"},
/* 463 */
    {ERR_NOPERMFORHOST, ":Tu host no esta privilegiado"},
/* 464 */
    {ERR_PASSWDMISMATCH, ":Pass incorrecta"},
/* 465 */
    {ERR_YOUREBANNEDCREEP, ":Estas baneado de este servidor"},
/* 466 */
    {ERR_YOUWILLBEBANNED, (char *)NULL},
/* 467 */
    {ERR_KEYSET, "%s :La pass del canal ya esta fijada"},
/* 468 */
    {ERR_INVALIDUSERNAME, (char *)NULL},
/* 469 */
    {0, (char *)NULL},
/* 470 */
    {ERR_NEEDOPERSTATUS, "%s :No puedes entrar al canal (+A)"},
/* 471 */
    {ERR_CHANNELISFULL, "%s :No puedes entrar al canal (+l)"},
/* 472 */
    {ERR_UNKNOWNMODE, "%c :no conozco ese modo"},
/* 473 */
    {ERR_INVITEONLYCHAN, "%s :No puedes entrar al canal (+i)"},
/* 474 */
    {ERR_BANNEDFROMCHAN, "%s :No puedes entrar al canal (+b)"},
/* 475 */
    {ERR_BADCHANNELKEY, "%s :No puedes entrar al canal (+k)"},
/* 476 */
    {ERR_BADCHANMASK, "%s :Mascara de canal invalida"},
/* 477 */
    {ERR_NEEDREGGEDNICK, "%s :No puedes entrar al canal (+R)"},
/* 478 */
    {ERR_BANLISTFULL, "%s %s :La lista de ignore esta llena"},
/* 479 */
    {ERR_BADCHANNAME, "%s :No puedes entrar al canal (acceso denegado en este servidor)"},
/* 480 */
    {0, (char *)NULL},
/* 481 */
    {ERR_NOPRIVILEGES, ":No tienes privilegios para ejecutar el comando"},
/* 482 */
    {ERR_CHANOPRIVSNEEDED, "%s :No eres operador del canal"},
/* 483 */
    {ERR_CANTKILLSERVER, ":No puedes matar un servidor!"},
/* 484 */
    {ERR_ISCHANSERVICE, "%s %s :No puedes echar o quitarle la @ un operador de la red"},
/* 485 */
    {0, (char *)NULL},
/* 486 */
    {ERR_NONONREG, ":Necesitas el modo +r para mandar un privado a este usuario %s"},
/* 487 */
    {0, (char *)NULL},
/* 488 */
    {0, (char *)NULL},
/* 489 */
    {ERR_VOICENEEDED, "%s :No tienes voz o arroba en el canal"},
/* 490 */
    {0, (char *)NULL},
/* 491 */
    {ERR_NOOPERHOST, ":No hay oline para tu host"},
/* 492 */
    {ERR_OPERCLASSFULL, ":Clase %d esta llena!"},
/* 493 */
    {ERR_NOOPERCLASS, ":Clase %d no existe"},
/* 494 */
    {0, (char *)NULL},
/* 495 */
    {0, (char *)NULL},
/* 496 */
    {0, (char *)NULL},
/* 497 */
    {0, (char *)NULL},
/* 498 */
    {ERR_ISOPERLCHAN, "%s %s :No puedes echar o quitarle la @ a un operador de la red"},
/* 499 */
    {0, (char *)NULL},
/* 500 */
    {0, (char *)NULL},
/* 501 */
    {ERR_UMODEUNKNOWNFLAG, ":Modo desconocido"},
/* 502 */
    {ERR_USERSDONTMATCH, ":No puedes cambiar los modos de otros usuarios"},
/* 503 */
    {0, (char *)NULL},
/* 504 */
    {0, (char *)NULL},
/* 505 */
    {0, (char *)NULL},
/* 506 */
    {0, (char *)NULL},
/* 507 */
    {0, (char *)NULL},
/* 508 */
    {0, (char *)NULL},
/* 509 */
    {ERR_SILECANTBESHOWN, "%s :No puedes mirar la lista de silence de otro usuario"},
/* 510 */
    {ERR_ISSILENCING, ":No puedes hablar con %s. Te ha silenciado"},
/* 511 */
    {ERR_SILELISTFULL, "%s :Tu lista de silence esta llena"},
#if defined(WATCH)
    {ERR_TOOMANYWATCH, "%s :El maximo para tu lista de WATCH es de %d entradas"},
#else
/* 512 */
    {0, (char *)NULL},
#endif /* WATCH */
/* 513 */
    {ERR_BADPING, (char *)NULL},
/* 514 */
    {ERR_NOSUCHGLINE, "%s@%s :No encuentro gline"}
};

static Numeric numeric_replies[] = {
/* 300 */
    {RPL_NONE, (char *)NULL},
/* 301 */
    {RPL_AWAY, "%s :%s"},
/* 302 */
    {RPL_USERHOST, ":"},
/* 303 */
    {RPL_ISON, ":"},
/* 304 */
    {RPL_TEXT, (char *)NULL},
/* 305 */
    {RPL_UNAWAY, ":Has vuelto del away, buen chat"},
/* 306 */
    {RPL_NOWAWAY, ":Has sido marcado como ausente"},
/* 307 */
    {RPL_WHOISREGNICK, "%s :Tiene el nick Registrado y Protegido"},
/* 308 */
     {RPL_WHOISDEVEL, "%s :Es un DeVeLoper de %s"},
/* 309 */
    {RPL_WHOISADMIN, "%s :Es un ADMINistrador de %s"},
/* 310 */
    {RPL_WHOISHELPOP, "%s :Es un OPERador de %s"},
/* 311 */
    {RPL_WHOISUSER, "%s %s %s * :%s"},
/* 312 */
    {RPL_WHOISSERVER, "%s %s :%s"},
/* 313 */
    {RPL_WHOISOPERATOR, "%s :Es un IRCOP de %s"},
/* 314 */
    {RPL_WHOWASUSER, "%s %s %s * :%s"},
/* 315 */
    {RPL_ENDOFWHO, "%s :Fin del /WHO"},
/* 316 */
    {RPL_WHOISBOT, "%s :Es un roBOT oficial de %s"},
/* 317 */
    {RPL_WHOISIDLE, "%s %ld %ld :segundos idle, online desde"},
/* 318 */
    {RPL_ENDOFWHOIS, "%s :Fin del /WHOIS list"},
/* 319 */
    {RPL_WHOISCHANNELS, "%s :%s"},
/* 320 */
    {RPL_WHOISPREOPER, "%s :Es un PRE-OPERador de %s"},
/* 321 */
    {RPL_LISTSTART, "Canal :Usuarios  Nombre"},
/* 322 */
    {RPL_LIST, "%s %d :%s"},
/* 323 */
    {RPL_LISTEND, ":Fin del /LIST"},
/* 324 */
    {RPL_CHANNELMODEIS, "%s %s %s"},
/* 325 */
    {RPL_HIDDENWHOIS, "%s :Esta Escondido"},  
/* 326 */
    {RPL_WHOISRBT, "%s :Es un roBOT oficial de %s"},
/* 327 */
    {RPL_WHOISDJ, "%s :Es un Deejay de %s"},
/* 328 */
    {0, (char *)NULL},
/* 329 */
    {RPL_CREATIONTIME, "%s " TIME_T_FMT},
/* 330 */
    {0, (char *)NULL},
/* 331 */
    {RPL_NOTOPIC, "%s :No hay topic definido"},
/* 332 */
    {RPL_TOPIC, "%s :%s"},
/* 333 */
    {RPL_TOPICWHOTIME, "%s %s " TIME_T_FMT},
/* 334 */
    {RPL_LISTUSAGE, ":%s"},
/* 335 */
    {0, (char *)NULL},
/* 336 */
    {0, (char *)NULL},
/* 337 */
    {0, (char *)NULL},
/* 338 */
    {0, (char *)NULL},
/* 339 */
    {0, (char *)NULL},
/* 340 */
    {RPL_USERIP, ":"},
/* 341 */
    {RPL_INVITING, "%s %s"},
/* 342 */
    {RPL_MSGONLYREG, "%s :Solo admite privados de usuarios registrados"},
/* 343 */
    {0, (char *)NULL},
/* 344 */
    {0, (char *)NULL},
/* 345 */
    {0, (char *)NULL},
/* 346 */
    {RPL_INVITELIST, ":%s"},
/* 347 */
    {RPL_ENDOFINVITELIST, ":Fin de la lista de invitados"},
/* 348 */
    {0, (char *)NULL},
/* 349 */
    {0, (char *)NULL},
/* 350 */
    {0, (char *)NULL},
/* 351 */
    {RPL_VERSION, "%s.%s %s :%s"},
/* 352 */
    {RPL_WHOREPLY, "%s"},
/* 353 */
    {RPL_NAMREPLY, "%s"},
/* 354 */
    {RPL_WHOSPCRPL, "%s"},
/* 355 */
    {0, (char *)NULL},
/* 356 */
    {0, (char *)NULL},
/* 357 */
    {0, (char *)NULL},
/* 358 */
    {0, (char *)NULL},
/* 359 */
    {0, (char *)NULL},
/* 360 */
    {0, (char *)NULL},
/* 361 */
    {RPL_KILLDONE, (char *)NULL},
/* 362 */
    {RPL_CLOSING, "%s :Closed. Status = %d"},
/* 363 */
    {RPL_CLOSEEND, "%d: Connections Closed"},
/* 364 */
#if !defined(GODMODE)
    {RPL_LINKS, "%s %s :%d P%u %s"},
#else /* GODMODE */
    {RPL_LINKS, "%s %s :%d P%u " TIME_T_FMT " (%s) %s"},
#endif /* GODMODE */
/* 365 */
    {RPL_ENDOFLINKS, "%s :Fin del /LINKS list"},
/* 366 */
    {RPL_ENDOFNAMES, "%s :Fin del /NAMES list"},
/* 367 */
    {RPL_BANLIST, "%s %s %s " TIME_T_FMT},
/* 368 */
    {RPL_ENDOFBANLIST, "%s :Fin del Channel Ban List"},
/* 369 */
    {RPL_ENDOFWHOWAS, "%s :Fin del WHOWAS"},
/* 370 */
    {0, (char *)NULL},
/* 371 */
    {RPL_INFO, ":%s"},
/* 372 */
    {RPL_MOTD, ":- %s"},
/* 373 */
    {RPL_INFOSTART, ":Server INFO"},
/* 374 */
    {RPL_ENDOFINFO, ":Fin del /INFO list"},
/* 375 */
    {RPL_MOTDSTART, ":- %s Mensaje del dia - "},
/* 376 */
    {RPL_ENDOFMOTD, ":Fin del /MOTD command"},
/* 377 */
    {0, (char *)NULL},
/* 378 */
    {RPL_WHOISHOST, "%s :Direccin VIRTUAL %s"},
/* 379 */
    {RPL_WHOISMODES, "%s :Utiliza los modos [%s]"},
/* 380 */
    {0, (char *)NULL},
/* 381 */
    {RPL_YOUREOPER, ":Es ahora un IRCOP"},
/* 382 */
    {RPL_REHASHING, "%s :Rehashing"},
/* 383 */
    {0, (char *)NULL},
/* 384 */
    {RPL_MYPORTIS, "%d :Port to local server is\r\n"},
/* 385 */
    {RPL_NOTOPERANYMORE, (char *)NULL},
/* 386 */
    {RPL_YOUREADMIN, ":Es ahora un administrador"},
/* 387 */
    {RPL_YOUREHELP, ":Es ahora un operador"},
/* 388 */
    {RPL_YOUREDEVEL, ":Es ahora un desarrollador"},
/* 389 */
    {RPL_YOUREPREO, ":Es ahora un preoperador"},
/* 390 */
    {RPL_WHOISSUSPNICK, "%s :Tiene el nick SUSPENDido"},
/* 391 */
    {RPL_TIME, "%s " TIME_T_FMT " %ld :%s"},
/* 392 */
    {RPL_YOUREDJ, ":Es ahora un Deejay"},
/* 393 */
    {0, (char *)NULL},
/* 394 */
    {0, (char *)NULL},
/* 395 */
    {0, (char *)NULL},
/* 396 */
    {RPL_HOSTHIDDEN, "%s es tu nueva ip virtual"},
/* 397 */
    {0, (char *)NULL},
/* 398 */
    {0, (char *)NULL},
/* 399 */
    {0, (char *)NULL},
/* 200 */
#if !defined(GODMODE)
    {RPL_TRACELINK, "Link %s%s %s %s"},
#else /* GODMODE */
    {RPL_TRACELINK, "Link %s%s %s %s " TIME_T_FMT},
#endif /* GODMODE */
/* 201 */
    {RPL_TRACECONNECTING, "Try. %d %s"},
/* 202 */
    {RPL_TRACEHANDSHAKE, "H.S. %d %s"},
/* 203 */
    {RPL_TRACEUNKNOWN, "???? %d %s"},
/* 204 */
    {RPL_TRACEOPERATOR, "Oper %d %s %ld"},
/* 205 */
    {RPL_TRACEUSER, "User %d %s %ld"},
/* 206 */
    {RPL_TRACESERVER, "Serv %d %dS %dC %s %s!%s@%s %ld %ld"},
/* 207 */
    {0, (char *)NULL},
/* 208 */
    {RPL_TRACENEWTYPE, "<newtype> 0 %s"},
/* 209 */
    {RPL_TRACECLASS, "Class %d %d"},
/* 210 */
    {0, (char *)NULL},
/* 211 */
    {RPL_STATSLINKINFO, (char *)NULL},
/* 212 */
    {RPL_STATSCOMMANDS, "%s %u %u"},
/* 213 */
    {RPL_STATSCLINE, "%c %s * %s %d %d"},
/* 214 */
    {RPL_STATSNLINE, "%c %s * %s %d %d"},
/* 215 */
    {RPL_STATSILINE, "%c %s * %s %d %d"},
/* 216 */
    {RPL_STATSKLINE, "%c %s %s %s %d %d"},
/* 217 */
    {RPL_STATSPLINE, "%c %d %d %#x"},
/* 218 */
    {RPL_STATSYLINE, "%c %d %d %d %d (%d) %ld"},
/* 219 */
    {RPL_ENDOFSTATS, "%c :Fin del /STATS"},
/* 220 */
    {0, (char *)NULL},
/* 221 */
    {RPL_UMODEIS, "%s"},
/* 222 */
    {0, (char *)NULL},
/* 223 */
    {RPL_STATSELINE, "%c %s * %s %d %d"},
/* 224 */
    {0, (char *)NULL},
/* 225 */
    {0, (char *)NULL},
/* 226 */
    {0, (char *)NULL},
/* 227 */
    {0, (char *)NULL},
/* 228 */
    {0, (char *)NULL},
/* 229 */
    {0, (char *)NULL},
/* 230 */
    {0, (char *)NULL},
/* 231 */
    {RPL_SERVICEINFO, (char *)NULL},
/* 232 */
    {RPL_ENDOFSERVICES, (char *)NULL},
/* 233 */
    {RPL_SERVICE, (char *)NULL},
/* 234 */
    {RPL_SERVLIST, (char *)NULL},
/* 235 */
    {RPL_SERVLISTEND, (char *)NULL},
/* 236 */
    {0, (char *)NULL},
/* 237 */
    {0, (char *)NULL},
/* 238 */
    {0, (char *)NULL},
/* 239 */
    {0, (char *)NULL},
/* 240 */
    {0, (char *)NULL},
/* 241 */
    {RPL_STATSLLINE, "%c %s * %s %d %d"},
/* 242 */
    {RPL_STATSUPTIME, ":Servidor encendido hace %d dias, %d:%02d:%02d"},
/* 243 */
    {RPL_STATSOLINE, "%c %s * %s %d %d"},
/* 244 */
    {RPL_STATSHLINE, "%c %s * %s %d %d"},
/* 245 */
    {0, (char *)NULL},
/* 246 */
    {RPL_STATSTLINE, "%c %s %s"},
/* 247 */
    {RPL_STATSGLINE, "%c %s@%s " TIME_T_FMT " :%s (expira a %s)"},
/* 248 */
    {RPL_STATSULINE, "%c %s %s %s %d %d"},
/* 249 */
    {0, (char *)NULL},
/* 250 */
    {RPL_STATSCONN, ":Maximo de conexiones: %d (%d clientes)"},
/* 251 */
    {RPL_LUSERCLIENT, ":Aqu hay %d usuarios y %d invisibles en %d servidores"},
/* 252 */
    {RPL_LUSEROP, "%d :Operador(es), %d IRCop(s) , %d Bot(s) oficial(es) y %d Administrador(es)"},
/* 253 */
    {RPL_LUSERUNKNOWN, "%d :conexiones desconocida(s)"},
/* 254 */
    {RPL_LUSERCHANNELS, "%d :canales creados"},
/* 255 */
    {RPL_LUSERME, ":Aqu tenemos %d clientes y %d servidores"},
/* 256 */
    {RPL_ADMINME, ":Informacion administrativa sobre %s"},
/* 257 */
    {RPL_ADMINLOC1, ":%s"},
/* 258 */
    {RPL_ADMINLOC2, ":%s"},
/* 259 */
    {RPL_ADMINEMAIL, ":%s"},
/* 260 */
    {0, (char *)NULL},
/* 261 */
    {RPL_TRACELOG, "File %s %d"},
/* 262 */
    {RPL_TRACEPING, "Ping %s %s"},
/* 263 */
    {0, (char *)NULL},
/* 264 */
    {0, (char *)NULL},
/* 265 */ /* Dal.Net: Cadena numerica de usuarios locales */
    {RPL_LOCALUSERS, ":Usuarios locales actuales: %d Maximo: %d (%s, desde %s)"},
/* 266 */ /* Dal.Net: Cadena numerica de usuarios globales */
    {RPL_GLOBALUSERS, ":Usuarios globales actuales: %d Maximo: %d (%s, desde %s)"},
/* 267 */
    {0, (char *)NULL},
/* 268 */
    {0, (char *)NULL},
/* 269 */
    {0, (char *)NULL},
/* 270 */
    {0, (char *)NULL},
/* 271 */
    {RPL_SILELIST, "%s %s"},
/* 272 */
    {RPL_ENDOFSILELIST, "%s :Fin de la lista de silence"},
/* 273 */
    {0, (char *)NULL},
/* 274 */
    {0, (char *)NULL},
/* 275 */
    {RPL_STATSDLINE, "%c %s %s"},
/* 276 */
    {0, (char *)NULL},
/* 277 */
    {0, (char *)NULL},
/* 278 */
    {0, (char *)NULL},
/* 279 */
    {0, (char *)NULL},
/* 280 */
    {RPL_GLIST, "%s@%s " TIME_T_FMT " %s%s (expira a %s)"},
/* 281 */
    {RPL_ENDOFGLIST, ":Fin de la lista de G-line"}
};


#if defined(WATCH)
/*
 * Numericos de watch
 */
static Numeric watch_replies[] = {
/* 600 */
    {RPL_LOGON, "%s %s %s %ld :se conecta"},
/* 601 */
    {RPL_LOGOFF, "%s %s %s %ld :se desconecta"},
/* 602 */
    {RPL_WATCHOFF, "%s %s %s %ld :stopped watching"},
/* 603 */
#if 0  /* Desactivado */
    {RPL_WATCHSTAT, ":You have %d and are on %d WATCH entries"},
#else
    {RPL_WATCHSTAT, ":You have %d WATCH entries"},
#endif
/* 604 */
    {RPL_NOWON, "%s %s %s %ld :esta en linea"},
/* 605 */
    {RPL_NOWOFF, "%s %s %s %ld :no esta conectado"},
/* 606 */
    {RPL_WATCHLIST, ":%s"},
/* 607 */
    {RPL_ENDOFWATCHLIST, ":Fin del WATCH %c"}
};
#endif /* WATCH */


/* *INDENT-ON* */

static char numbuff[512];

/* "inline" */
#define prepbuf(buffer, num, tail)			\
{							\
  register char *s = buffer + 4;			\
  register const char *ap = atoi_tab + (num << 2);	\
  							\
  strcpy(buffer, ":%s 000 %s ");			\
  *s++ = *ap++;						\
  *s++ = *ap++;						\
  *s = *ap;						\
  strcpy(s + 5, tail);					\
}

char *err_str(int numeric)
{
  Reg1 Numeric *nptr;
  Reg2 int num = numeric;

  num -= numeric_errors[0].num_val;

#if defined(DEBUGMODE)
  if (num < 0 || num > ERR_USERSDONTMATCH)
    sprintf_irc(numbuff,
        ":%%s %d %%s :INTERNAL ERROR: BAD NUMERIC! %d", numeric, num);
  else
  {
    nptr = &numeric_errors[num];
    if (!nptr->num_form || !nptr->num_val)
      sprintf_irc(numbuff, ":%%s %d %%s :NO ERROR FOR NUMERIC ERROR %d",
          numeric, num);
    else
      prepbuf(numbuff, nptr->num_val, nptr->num_form);
  }
#else
  nptr = &numeric_errors[num];
  prepbuf(numbuff, nptr->num_val, nptr->num_form);
#endif

  return numbuff;
}

char *rpl_str(int numeric)
{
  Reg1 Numeric *nptr;
  Reg2 int num = numeric;

  if (num > (int)(sizeof(local_replies) / sizeof(Numeric) - 2))
    num -= (num > 300) ? 300 : 100;

#if defined(DEBUGMODE)
  if (num < 0 || num > 200)
    sprintf_irc(numbuff, ":%%s %d %%s :INTERNAL REPLY ERROR: BAD NUMERIC! %d",
        numeric, num);
  else
  {
    if (numeric > 99)
      nptr = &numeric_replies[num];
    else
      nptr = &local_replies[num];
    Debug((DEBUG_NUM, "rpl_str: numeric %d num %d nptr %p %d %p",
        numeric, num, nptr, nptr->num_val, nptr->num_form));
    if (!nptr->num_form || !nptr->num_val)
      sprintf_irc(numbuff, ":%%s %d %%s :NO REPLY FOR NUMERIC ERROR %d",
          numeric, num);
    else
      prepbuf(numbuff, nptr->num_val, nptr->num_form);
  }
#else
  if (numeric > 99)
    nptr = &numeric_replies[num];
  else
    nptr = &local_replies[num];
  prepbuf(numbuff, nptr->num_val, nptr->num_form);
#endif

  return numbuff;
}

#if defined(WATCH)
char *watch_str(int numeric)
{
  Reg1 Numeric *nptr;
  Reg2 int num = numeric;

  num -= watch_replies[0].num_val;

#if defined(DEBUGMODE)
  if (num < 0 || num > 7)       /* 600 - 607 */
    sprintf_irc(numbuff,
        ":%%s %d %%s :INTERNAL ERROR: BAD NUMERIC! %d", numeric, num);
  else
  {
    nptr = &watch_replies[num];
    if (!nptr->num_form || !nptr->num_val)
      sprintf_irc(numbuff, ":%%s %d %%s :NO ERROR FOR NUMERIC ERROR %d",
          numeric, num);
    else
      prepbuf(numbuff, nptr->num_val, nptr->num_form);
  }
#else
  nptr = &watch_replies[num];
  prepbuf(numbuff, nptr->num_val, nptr->num_form);
#endif

  return numbuff;
}

#endif /* WATCH */
