/*
 * Red-Unida IRC Daemon
 * Copyright (C) 2004-2005 Micael Fernandez
 *
 *
 * Basado sobre el codigo de Jarkko Oikarinen, UnderNet y IRC-Hispano.
 *
 * Este programa es gratis, puedes redistribuirlo y modificarlo bajo los
 * terminos de la GNU.
 *
 */


#include "sys.h"

#if defined(ESNET_NEG)

#include "h.h"
#include "s_debug.h"
#include "common.h"
#include "ircd.h"
#include "struct.h"
#include "s_serv.h"
#include "msg.h"
#include "runmalloc.h"
#include "s_conf.h"
#include "m_config.h"
#include "send.h"

#include <assert.h>

typedef enum { FIN, ZLIB } opcion;

static struct {
  char *texto;
  opcion opcion;
} opciones[] = {
#if defined(ZLIB_ESNET)
  {
  "zlib", ZLIB},
#endif
  {
NULL, FIN}};

#if defined(ZLIB_ESNET)
voidpf z_alloc(voidpf opaque, uInt items, uInt size)
{
  return RunCalloc(items, size);
}

void z_free(voidpf opaque, voidpf address)
{
  RunFree(address);
}

#endif

typedef enum { YO2EL, EL2YO } tipo_config;

static char *mira_conf_negociacion(char *nombre, tipo_config tipo_config)
{
  aConfItem *p;

  for (p = conf_negociacion; p; p = p->next)
  {
    if (!strCasediff(nombre, p->name))
      break;
  }
  if (!p)
    return "";
  if (tipo_config == YO2EL)
    return p->host;
  return p->passwd;
}

void config_resolve_speculative(aClient *cptr)
{
  char *p;
  int estado;

  p = mira_conf_negociacion(cptr->name, YO2EL);

  if (IsServer(cptr))
  {
#if defined(ZLIB_ESNET)
    if (!strchr(p, 'z') && (cptr->negociacion & ZLIB_ESNET_OUT_SPECULATIVE))
    {
      sendto_one(cptr, ":%s " MSG_CONFIG " ACK :zlib", me.name);
      cptr->negociacion &= ~ZLIB_ESNET_OUT_SPECULATIVE;
      cptr->negociacion |= ZLIB_ESNET_OUT;
      cptr->comp_out = RunMalloc(sizeof(z_stream));
      if (!cptr->comp_out)
        outofmemory();
      cptr->comp_out->zalloc = z_alloc;
      cptr->comp_out->zfree = z_free;
      cptr->comp_out->opaque = 0;
      estado = deflateInit(cptr->comp_out, 9);
      assert(estado == Z_OK);
      cptr->comp_out_total_in = 0;
      cptr->comp_out_total_out = 0;
    }
#endif
  }                             /* No es server */
}

int config_ack(aClient *cptr, aClient *sptr, char *fuente, char *valor,
    opcion opcion)
{
  int estado;

  if (IsServer(cptr))
  {
    switch (opcion)
    {
#if defined(ZLIB_ESNET)
      case ZLIB:
        cptr->negociacion |= ZLIB_ESNET_IN;
        cptr->comp_in = RunMalloc(sizeof(z_stream));
        if (!cptr->comp_in)
          outofmemory();
        cptr->comp_in->next_in = Z_NULL;
        cptr->comp_in->avail_in = 0;
        cptr->comp_in->zalloc = z_alloc;
        cptr->comp_in->zfree = z_free;
        cptr->comp_in->opaque = 0;
        estado = inflateInit(cptr->comp_in);
        assert(estado == Z_OK);
        cptr->comp_in_total_in = 0;
        cptr->comp_in_total_out = 0;
        break;
#endif
      default:                 /* No deberia ocurrir nunca */
        break;
    }
  }
  return 0;
}

int config_req(aClient *cptr, aClient *sptr, char *fuente, char *valor,
    opcion opcion)
{
  char *p;

  p = mira_conf_negociacion(PunteroACadena(sptr->name), YO2EL);

  if (IsServer(sptr))
  {
    switch (opcion)
    {
#if defined(ZLIB_ESNET)
      case ZLIB:
        if (!strchr(p, 'z'))
        {                       /* Permitimos compresion */
          sendto_one(sptr, ":%s " MSG_CONFIG " ACK :zlib", me.name);
          cptr->negociacion |= ZLIB_ESNET_OUT;
        }
        break;
#endif
      case FIN:
        assert(0);              /* No deberia darse nunca */
    }
  }
  else
  {                             /* No es servidor o no lo sabemos aun */
    switch (opcion)
    {
#if defined(ZLIB_ESNET)
      case ZLIB:
        cptr->negociacion |= ZLIB_ESNET_OUT_SPECULATIVE;
        break;
#endif
      case FIN:
        assert(0);              /* No deberia darse nunca */
    }
  }
  return 0;
}

int m_config(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
  int i;

  if (!MyConnect(cptr) || (cptr != sptr))
  {
    sendto_one(sptr, ":%s NOTICE %s :Solo conexiones locales", me.name,
        parv[0]);
    return 0;
  }
  if (parc < 3)
  {
    return 0;
  }

  for (i = 0; opciones[i].texto != NULL; i++)
  {
    if (!strcasecmp(opciones[i].texto, parv[2]))
      break;
  }
  if (opciones[i].texto == NULL)
  {
    sendto_one(sptr, ":%s NOTICE %s :Valor negociado desconocido", me.name,
        PunteroACadena(sptr->name));
    return 0;
  }

  if (!strcasecmp("ACK", parv[1]))
    return config_ack(cptr, sptr, parv[0], parv[2], opciones[i].opcion);
  if (!strcasecmp("REQ", parv[1]))
    return config_req(cptr, sptr, parv[0], parv[2], opciones[i].opcion);

  return 0;
}

void envia_config_req(aClient *cptr)
{
  char *p;

  p = mira_conf_negociacion(cptr->name, EL2YO);
#if defined(ZLIB_ESNET)
  if (!strchr(p, 'z'))          /* Queremos que nos mande comprimido */
    sendto_one(cptr, ":%s " MSG_CONFIG " REQ :zlib", me.name);
#endif
}

#endif /* M_CONFIG.C */
